from json import dumps as json_dumps
from typing import Any, Callable, Dict, List, Sequence, Type, Union, cast
from unittest.mock import Mock
from urllib.parse import urlencode

from ninja import NinjaAPI, Router
from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.testing.client import NinjaClientBase, NinjaResponse
from ninja_extra import ControllerBase

from easy.main import EasyAPI


class EasyAPIClientBase(NinjaClientBase):
    def __init__(
        self,
        router_or_app: Union[EasyAPI, Router, Type[ControllerBase]],
        auth: Union[
            Sequence[Callable[..., Any]], Callable[..., Any], NOT_SET_TYPE, None
        ] = NOT_SET,
        api_cls: Union[Type[EasyAPI], Type] = EasyAPI,
    ) -> None:
        if hasattr(router_or_app, "get_api_controller"):
            api = api_cls(auth=auth)
            controller_ninja_api_controller = router_or_app.get_api_controller()  # type: ignore
            assert controller_ninja_api_controller
            controller_ninja_api_controller.set_api_instance(api)
            self._urls_cache = list(controller_ninja_api_controller.urls_paths(""))
            router_or_app = api
        super().__init__(cast(Union[NinjaAPI, Router], router_or_app))

    def request(
        self,
        method: str,
        path: str,
        data: Dict = {},
        json: Any = None,
        **request_params: Any,
    ) -> "NinjaResponse":
        if json is not None:
            request_params["body"] = json_dumps(json)
        if "query" in request_params and isinstance(request_params["query"], dict):
            query = request_params.pop("query")
            url_encode = urlencode(query)
            path = f"{path}?{url_encode}"
        func, request, kwargs = self._resolve(method, path, data, request_params)
        return self._call(func, request, kwargs)  # type: ignore

    @property
    def urls(self) -> List:
        if not hasattr(self, "_urls_cache"):
            self._urls_cache: List
            if isinstance(self.router_or_app, EasyAPI):
                self._urls_cache = self.router_or_app.urls[0]  # pragma: no cover
            else:
                api = EasyAPI()
                self.router_or_app.set_api_instance(api)  # type: ignore
                self._urls_cache = list(self.router_or_app.urls_paths(""))  # type: ignore
        return self._urls_cache


class EasyTestClient(EasyAPIClientBase):
    async def _call(self, func: Callable, request: Mock, kwargs: Dict) -> NinjaResponse:
        return NinjaResponse(await func(request, **kwargs))



from easy.main import EasyAPI

from .auth import jwt_auth_async
from .controllers import (
    AutoGenCrudAPIController,
    EasyCrudAPIController,
    PermissionAPIController,
)

api_unittest = EasyAPI(auth=jwt_auth_async)
api_unittest.register_controllers(
    EasyCrudAPIController,
    PermissionAPIController,
    AutoGenCrudAPIController,
)



import logging

from easy.controller.meta import CrudApiMetaclass

logger = logging.getLogger(__name__)


class CrudAPIController(metaclass=CrudApiMetaclass):
    """
    Base APIController for auto creating CRUD APIs, configurable via Meta class
    APIs auto generated:
        Creat
            PUT /{id}       - Create a single Object

        Read
            GET /{id}       - Retrieve a single Object
            GET /           - Retrieve multiple Object, paginated, support filtering

        Update
            PATCH /{id}     - Update a single Object

        Delete
            DELETE /{id}    - Delete a single Object

    Configuration:
        model:              django model
        generate_crud:      whether to create crud api, default to True
        model_exclude:      fields to be excluded in Schema, it will ignore model_fields
        model_fields:       fields to be included in Schema, default to "__all__"
        model_join:         retrieve all m2m fields, default to True
        model_recursive:    recursively retrieve FK/OneToOne models, default to False
        sensitive_fields:   fields to be ignored

    Example:
        class Meta
            model = Event
            generate_crud = False
            model_exclude = ["field1", "field2"]
            model_fields = ["field1", "field2"]
            model_join = False
            model_recursive = True
            sensitive_fields = ["token", "money"]
    """

    ...



from django.contrib.auth import get_user_model
from factory import Faker
from factory.django import DjangoModelFactory


class UserFactory(DjangoModelFactory):

    username = Faker("user_name")
    email = Faker("email")

    class Meta:
        model = get_user_model()
        django_get_or_create = ["username"]



from functools import wraps
from types import FunctionType
from typing import Any, Callable
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http import HttpRequest
from django.shortcuts import resolve_url


def request_passes_test(
    test_func: Callable[[Any], Any],
    login_url: str = None,
    redirect_field_name: str = REDIRECT_FIELD_NAME,
) -> Callable[[FunctionType], Callable[[HttpRequest, Any], Any]]:
    """
    Decorator for views that checks that the request passes the given test,
    redirecting to the log-in page if necessary. The test should be a callable
    that takes the user object and returns True if the user passes.
    """

    def decorator(view_func: FunctionType) -> Callable[[HttpRequest, Any], Any]:
        @wraps(view_func)
        def _wrapped_view(request: HttpRequest, *args: Any, **kwargs: Any) -> Any:
            if test_func(request):
                return view_func(request, *args, **kwargs)
            path = request.build_absolute_uri()
            resolved_login_url = resolve_url(login_url or settings.LOGIN_URL)
            # If the login url is the same scheme and net location then just
            # use the path as the "next" url.
            login_scheme, login_netloc = urlparse(resolved_login_url)[:2]
            current_scheme, current_netloc = urlparse(path)[:2]
            if (not login_scheme or login_scheme == current_scheme) and (
                not login_netloc or login_netloc == current_netloc
            ):
                path = request.get_full_path()
            from django.contrib.auth.views import redirect_to_login

            return redirect_to_login(path, resolved_login_url, redirect_field_name)

        return _wrapped_view

    return decorator


def docs_permission_required(
    view_func: FunctionType = None,
    redirect_field_name: str = REDIRECT_FIELD_NAME,
    login_url: str = "admin:login",
) -> Any:
    """
    Decorator for views that checks that the user is logged in and is a staff
    member, redirecting to the login page if necessary.
    """
    actual_decorator = request_passes_test(
        lambda r: ((r.user.is_active and r.user.is_staff)),
        login_url=login_url,
        redirect_field_name=redirect_field_name,
    )
    if view_func:
        return actual_decorator(view_func)
    return actual_decorator  # pragma: no cover



import json
import logging
import re
import uuid
from abc import ABCMeta
from collections import ChainMap
from typing import Any, Match, Optional, Tuple, Type

from django.http import HttpRequest
from ninja import ModelSchema
from ninja_extra import ControllerBase, http_delete, http_get, http_patch, http_put
from ninja_extra.pagination import paginate

from easy.controller.meta_conf import MODEL_FIELDS_ATTR_DEFAULT, ModelOptions
from easy.domain.orm import CrudModel
from easy.response import BaseApiResponse
from easy.services import BaseService
from easy.utils import copy_func

logger = logging.getLogger(__name__)


class CrudAPI(CrudModel):
    # Never add type note to service, it will cause injection error
    def __init__(self, service=None):  # type: ignore
        # Critical to set __Meta
        self.service = service
        if self.service:
            self.model = self.service.model

        _model_opts: ModelOptions = ModelOptions.get_model_options(self.__class__)
        if self.model and _model_opts:
            ModelOptions.set_model_meta(self.model, _model_opts)

        if not service:
            self.service = BaseService(model=self.model)
        super().__init__(model=self.model)


class CrudApiMetaclass(ABCMeta):
    def __new__(mcs, name: str, bases: Tuple[Type[Any], ...], attrs: dict) -> Any:
        def is_private_attrs(attr_name: str) -> Optional[Match[str]]:
            return re.match(r"^__[^\d\W]\w*\Z__$", attr_name, re.UNICODE)

        parent_attrs = ChainMap(
            *[attrs]
            + [
                {k: v for (k, v) in vars(base).items() if not (is_private_attrs(k))}
                for base in bases
            ]
        )
        base_cls_attrs: dict = {}
        base_cls_attrs.update(parent_attrs)

        # Get configs from Meta
        _temp_cls: Type = super().__new__(mcs, name, (object,), base_cls_attrs)
        model_opts: ModelOptions = ModelOptions.get_model_options(_temp_cls)

        # Define Controller APIs for auto generation
        async def get_obj(self, request: HttpRequest, id: int) -> Any:  # type: ignore
            """
            GET /{id}
            Retrieve a single Object
            """
            try:
                qs = await self.service.get_obj(id)
            except Exception as e:  # pragma: no cover
                logger.error(f"Get Error - {e}", exc_info=True)
                return BaseApiResponse(str(e), message="Get Failed", errno=500)
            if qs:
                return qs
            else:
                return BaseApiResponse(message="Not Found", errno=404)

        async def del_obj(self, request: HttpRequest, id: int) -> Any:  # type: ignore
            """
            DELETE /{id}
            Delete a single Object
            """
            if await self.service.del_obj(id):
                return BaseApiResponse("Deleted.", errno=204)
            else:
                return BaseApiResponse("Not Found.", errno=404)

        @paginate
        async def get_objs(self, request: HttpRequest, filters: str = None) -> Any:  # type: ignore
            """
            GET /?maximum={int}&filters={filters_dict}
            Retrieve multiple Object (optional: maximum # and filters)
            """
            if filters:
                return await self.service.get_objs(**json.loads(filters))
            return await self.service.get_objs()

        if model_opts.generate_crud and model_opts.model:
            base_cls_attrs.update(
                {
                    "get_obj": http_get("/{id}", summary="Get a single object")(
                        copy_func(get_obj)  # type: ignore
                    ),
                    "del_obj": http_delete("/{id}", summary="Delete a single object")(
                        copy_func(del_obj)  # type: ignore
                    ),
                    "get_objs": http_get("/", summary="Get multiple objects")(
                        copy_func(get_objs)  # type: ignore
                    ),
                }
            )

            class DataSchema(ModelSchema):
                class Config:
                    model = model_opts.model
                    if model_opts.model_exclude:
                        model_exclude = model_opts.model_exclude
                    else:
                        if model_opts.model_fields == MODEL_FIELDS_ATTR_DEFAULT:
                            model_fields = MODEL_FIELDS_ATTR_DEFAULT
                        else:
                            model_fields = (
                                model_opts.model_fields  # type: ignore
                                if model_opts.model_fields
                                else MODEL_FIELDS_ATTR_DEFAULT
                            )

            async def add_obj(  # type: ignore
                self, request: HttpRequest, data: DataSchema
            ) -> Any:
                """
                PUT /
                Create a single Object
                """
                obj_id = await self.service.add_obj(**data.dict())
                if obj_id:
                    return BaseApiResponse({"id": obj_id}, errno=201)
                else:
                    return BaseApiResponse("Add failed", errno=204)  # pragma: no cover

            async def patch_obj(  # type: ignore
                self, request: HttpRequest, id: int, data: DataSchema
            ) -> Any:
                """
                PATCH /{id}
                Update a single object
                """
                if await self.service.patch_obj(id=id, payload=data.dict()):
                    return BaseApiResponse("Updated.")
                else:
                    return BaseApiResponse("Update Failed", errno=400)

            DataSchema.__name__ = (
                f"{model_opts.model.__name__}__AutoSchema({str(uuid.uuid4())[:4]})"
            )

            base_cls_attrs.update(
                {
                    "patch_obj": http_patch("/{id}", summary="Patch a single object")(
                        copy_func(patch_obj)  # type: ignore
                    ),
                    "add_obj": http_put("/", summary="Create")(
                        copy_func(add_obj)  # type: ignore
                    ),
                }
            )

        new_cls: Type = super().__new__(
            mcs,
            name,
            (
                ControllerBase,
                CrudAPI,
            ),
            base_cls_attrs,
        )

        if model_opts.model:
            ModelOptions.set_model_meta(model_opts.model, model_opts)
            setattr(new_cls, "model", model_opts.model)

        return new_cls



import json
from datetime import datetime, timedelta

import django
import pytest
from asgiref.sync import sync_to_async

from .easy_app.controllers import (
    AutoGenCrudAPIController,
    AutoGenCrudNoJoinAPIController,
    AutoGenCrudSomeFieldsAPIController,
    EventSchema,
    InheritedRecursiveAPIController,
    NoCrudAPIController,
    RecursiveAPIController,
)
from .easy_app.models import Category, Client, Event, Type

dummy_data = dict(
    title="AsyncAdminAPIEvent",
    start_date=str(datetime.now().date()),
    end_date=str((datetime.now() + timedelta(days=5)).date()),
)


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
@pytest.mark.django_db
class TestAutoCrudAdminAPI:
    async def test_crud_generate_or_not(self, transactional_db, easy_api_client):
        client = easy_api_client(NoCrudAPIController)

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get")

        event = await sync_to_async(Event.objects.create)(**object_data)

        with pytest.raises(Exception):
            await client.get(
                f"/{event.id}",
            )

    async def test_crud_default_get_all(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get_all")

        event = await sync_to_async(Event.objects.create)(**object_data)
        type = await sync_to_async(Type.objects.create)(name="Type")
        category = await sync_to_async(Category.objects.create)(
            title="Category for Unit Testings"
        )
        client_a = await sync_to_async(Client.objects.create)(
            name="Client A for Unit Testings", key="A"
        )
        client_b = await sync_to_async(Client.objects.create)(
            name="Client B for Unit Testings", key="B"
        )
        event.category = category
        event.type = type
        await sync_to_async(event.save)()
        await sync_to_async(event.owner.set)([client_a, client_b])

        response = await client.get(
            "/", query=dict(maximum=100, filters=json.dumps(dict(id__gte=1)))
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"
        assert data[0]["type"] == type.id

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["start_date"] == data[0]["start_date"]

        # Recursive = True
        client = easy_api_client(RecursiveAPIController)
        response = await client.get(
            "/", query=dict(maximum=100, filters=json.dumps(dict(id__gte=1)))
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"

        assert data[0]["type"]["id"] == type.id
        assert data[0]["category"]["status"] == 1

        # Recursive = True, inherited class
        client = easy_api_client(InheritedRecursiveAPIController)
        response = await client.get(
            "/", query=dict(maximum=100, filters=json.dumps(dict(id__gte=1)))
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"

        assert data[0]["type"]["id"] == type.id
        assert data[0]["category"]["status"] == 1

        # Back to AutoGenCrudAPIController
        client = easy_api_client(AutoGenCrudAPIController)

        response = await client.get(
            "/",
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["start_date"] == data[0]["start_date"]

        response = await client.get("/")
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["start_date"] == data[0]["start_date"]

    async def test_crud_default_get_delete(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get")

        event = await sync_to_async(Event.objects.create)(**object_data)

        response = await client.get(
            f"/{event.id}",
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data["title"] == "AsyncAdminAPIEvent_get"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["end_date"] == data["end_date"]

        await client.delete(
            f"/{event.id}",
        )

        response = await client.get(
            f"/{event.id}",
        )
        assert response.status_code == 200
        assert response.json().get("code") == 404

    async def test_crud_default_create(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        client_c = await sync_to_async(Client.objects.create)(
            name="Client C for Unit Testings", key="C"
        )

        client_d = await sync_to_async(Client.objects.create)(
            name="Client D for Unit Testings", key="D"
        )

        type = await sync_to_async(Type.objects.create)(name="TypeForCreating")

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_create")
        object_data.update(owner=[client_c.id, client_d.id])
        object_data.update(lead_owner=[client_d.id])
        object_data.update(type_id=type.id)

        response = await client.put(
            "/", json=object_data, content_type="application/json"
        )
        assert response.status_code == 200

        assert response.json().get("code") == 201
        event_id = response.json().get("data")["id"]

        response = await client.get(
            f"/{event_id}",
        )

        assert response.status_code == 200
        assert response.json().get("data")["title"] == "AsyncAdminAPIEvent_create"

    async def test_crud_default_create_some_fields(
        self, transactional_db, easy_api_client
    ):
        client = easy_api_client(AutoGenCrudSomeFieldsAPIController)

        category = await sync_to_async(Category.objects.create)(
            title="Category for Unit Testings"
        )

        client_type = await sync_to_async(Client.objects.create)(
            name="Client for Unit Testings",
            key="Type",
            category=category,
            password="DUMMY_PASSWORD",
        )

        response = await client.get(
            f"/{client_type.id}",
        )

        assert response.status_code == 200
        assert response.json()["data"]["key"] == "Type"
        with pytest.raises(KeyError):
            print(response.json()["data"]["password"])
            print(response.json()["data"]["category"])

    async def test_crud_default_patch(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        object_data = dummy_data.copy()
        event = await sync_to_async(Event.objects.create)(**object_data)

        response = await client.get(
            f"/{event.pk}",
        )
        assert response.status_code == 200
        assert response.json().get("data")["title"] == f"{object_data['title']}"

        client_e = await sync_to_async(Client.objects.create)(
            name="Client E for Unit Testings", key="E"
        )

        client_f = await sync_to_async(Client.objects.create)(
            name="Client F for Unit Testings", key="F"
        )

        category = await sync_to_async(Category.objects.create)(
            title="Category for Unit Testings", status=2
        )

        new_data = dict(
            id=event.pk,
            title=f"{object_data['title']}_patch",
            start_date=str((datetime.now() + timedelta(days=10)).date()),
            end_date=str((datetime.now() + timedelta(days=20)).date()),
            owner=[client_e.pk, client_f.pk],
            category=category.pk,
        )

        response = await client.patch(
            "/20000", json=new_data, content_type="application/json"
        )

        assert response.status_code == 200
        assert response.json()["code"] == 400

        response = await client.patch(
            f"/{event.pk}", json=new_data, content_type="application/json"
        )

        assert response.status_code == 200
        assert response.json().get("data")

        response = await client.get(
            f"/{event.pk}",
        )
        assert response.status_code == 200
        data = response.json().get("data")

        assert len(data["owner"]) == 2
        assert len(data["lead_owner"]) == 0
        assert data["owner"][0]["name"] == "Client E for Unit Testings"
        assert data["owner"][1]["name"] == "Client F for Unit Testings"

        assert data["title"] == "AsyncAdminAPIEvent_patch"
        assert data["start_date"] == str((datetime.now() + timedelta(days=10)).date())
        assert data["end_date"] == str((datetime.now() + timedelta(days=20)).date())

        client = easy_api_client(AutoGenCrudNoJoinAPIController)

        # No auto join
        response = await client.get(
            f"/{event.pk}",
        )
        assert response.status_code == 200
        data = response.json().get("data")
        assert data["owner"] == [8, 9]



import logging
from typing import TYPE_CHECKING, Any

from django.http import HttpRequest

if TYPE_CHECKING:
    from ninja_extra.controllers.base import ControllerBase  # pragma: no cover


logger = logging.getLogger(__name__)


class PermissionService(object):
    """Base permission service for extra customization needs"""

    def check_permission(
        self, request: HttpRequest, controller: "ControllerBase"
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True

    def check_object_permission(
        self, request: HttpRequest, controller: "ControllerBase", obj: Any
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True



from django.test import RequestFactory
from django.utils.translation import gettext_lazy as _
from ninja_extra import exceptions

from easy import EasyAPI, testing
from easy.exception import APIAuthException, BaseAPIException

api = EasyAPI(urls_namespace="exception")


@api.get("/list_exception")
async def list_exception(request):
    raise BaseAPIException(
        [
            "some error 1",
            "some error 2",
        ]
    )


@api.get("/list_exception_full_detail")
async def list_exception_full(request):
    exception = BaseAPIException(
        [
            "some error 1",
            "some error 2",
        ]
    )
    return exception.get_full_details()


@api.get("/dict_exception")
async def dict_exception(request):
    raise BaseAPIException(dict(error="error 1"))


@api.get("/dict_exception_full_detail")
async def dict_exception_full_detail(request):
    exception = BaseAPIException(dict(error="error 1"))
    return exception.get_full_details()


@api.get("/dict_exception_code_detail")
async def dict_exception_code_detail(request):
    exception = BaseAPIException(dict(error="error 1"))
    return exception.get_codes()


@api.get("/list_exception_code_detail")
async def list_exception_code_detail(request):
    exception = BaseAPIException(["some error"])
    return exception.get_codes()


client = testing.EasyTestClient(api)


class TestException:
    async def test_get_error_details(self):
        example = "string"
        lazy_example = _(example)

        assert exceptions._get_error_details(lazy_example) == example

        assert isinstance(
            exceptions._get_error_details(lazy_example), exceptions.ErrorDetail
        )

        assert (
            exceptions._get_error_details({"nested": lazy_example})["nested"] == example
        )

        assert isinstance(
            exceptions._get_error_details({"nested": lazy_example})["nested"],
            exceptions.ErrorDetail,
        )

        assert exceptions._get_error_details([[lazy_example]])[0][0] == example

        assert isinstance(
            exceptions._get_error_details([[lazy_example]])[0][0],
            exceptions.ErrorDetail,
        )


class TestErrorDetail:
    async def test_eq(self):
        assert exceptions.ErrorDetail("msg") == exceptions.ErrorDetail("msg")
        assert exceptions.ErrorDetail("msg", "code") == exceptions.ErrorDetail(
            "msg", code="code"
        )

        assert exceptions.ErrorDetail("msg") == "msg"
        assert exceptions.ErrorDetail("msg", "code") == "msg"

    async def test_ne(self):
        assert exceptions.ErrorDetail("msg1") != exceptions.ErrorDetail("msg2")
        assert exceptions.ErrorDetail("msg") != exceptions.ErrorDetail(
            "msg", code="invalid"
        )

        assert exceptions.ErrorDetail("msg1") != "msg2"
        assert exceptions.ErrorDetail("msg1", "code") != "msg2"

    async def test_repr(self):
        assert repr(
            exceptions.ErrorDetail("msg1")
        ) == "ErrorDetail(string={!r}, code=None)".format("msg1")
        assert repr(
            exceptions.ErrorDetail("msg1", "code")
        ) == "ErrorDetail(string={!r}, code={!r})".format("msg1", "code")

    async def test_str(self):
        assert str(exceptions.ErrorDetail("msg1")) == "msg1"
        assert str(exceptions.ErrorDetail("msg1", "code")) == "msg1"

    async def test_hash(self):
        assert hash(exceptions.ErrorDetail("msg")) == hash("msg")
        assert hash(exceptions.ErrorDetail("msg", "code")) == hash("msg")


async def test_server_error():
    request = RequestFactory().get("/")
    response = exceptions.server_error(request)
    assert response.status_code == 500
    assert response["content-type"] == "application/json"


async def test_bad_request():
    request = RequestFactory().get("/")
    exception = Exception("Something went wrong — Not used")
    response = exceptions.bad_request(request, exception)
    assert response.status_code == 400
    assert response["content-type"] == "application/json"


async def test_exception_with_list_details():
    res = await client.get("list_exception")
    assert res.status_code == 500
    assert res.json()["data"] == {
        "detail": "[ErrorDetail(string='some error 1', code='error'), "
        "ErrorDetail(string='some error 2', code='error')]",
    }


async def test_exception_with_list_full_details():
    res = await client.get("list_exception_full_detail")
    assert res.status_code == 200
    assert res.json()["data"] == [
        {"message": "some error 1", "code": "error"},
        {"message": "some error 2", "code": "error"},
    ]


async def test_exception_with_dict_details():
    res = await client.get("dict_exception")
    assert res.status_code == 500
    assert res.json()["data"] == {
        "detail": "{'error': ErrorDetail(string='error 1', code='error')}"
    }


async def test_exception_with_full_details():
    res = await client.get("dict_exception_full_detail")
    assert res.status_code == 200
    assert res.json()["data"] == {"error": {"message": "error 1", "code": "error"}}


async def test_exception_dict_exception_code_detail():
    res = await client.get("dict_exception_code_detail")
    assert res.status_code == 200
    assert res.json()["data"] == {"error": "error"}


async def test_exception_with_list_exception_code_detail():
    res = await client.get("list_exception_code_detail")
    assert res.status_code == 200
    assert res.json()["data"] == ["error"]


def test_validation_error():
    exception = exceptions.ValidationError()
    assert exception.detail == [
        exceptions.ErrorDetail(
            string=exceptions.ValidationError.default_detail,
            code=exceptions.ValidationError.default_code,
        )
    ]
    assert exception.get_codes() == [exceptions.ValidationError.default_code]

    exception = exceptions.ValidationError(["errors"])
    assert exception.detail == ["errors"]


def test_method_not_allowed_error():
    exception = exceptions.MethodNotAllowed("get")
    assert exception.detail == exceptions.MethodNotAllowed.default_detail.format(
        method="get"
    )
    assert exception.get_codes() == exceptions.MethodNotAllowed.default_code

    exception = exceptions.MethodNotAllowed("get", ["errors"])
    assert exception.detail == ["errors"]


async def test_method_not_allowed_accepted_error():
    exception = exceptions.NotAcceptable()
    assert exception.detail == exceptions.NotAcceptable.default_detail
    assert exception.get_codes() == exceptions.NotAcceptable.default_code

    exception = exceptions.NotAcceptable(["errors"])
    assert exception.detail == ["errors"]


async def test_unsupported_media_type_allowed_error():
    exception = exceptions.UnsupportedMediaType("whatever/type")
    assert exception.detail == exceptions.UnsupportedMediaType.default_detail.format(
        media_type="whatever/type"
    )
    assert exception.get_codes() == exceptions.UnsupportedMediaType.default_code

    exception = exceptions.UnsupportedMediaType("whatever/type", ["errors"])
    assert exception.detail == ["errors"]


async def test_api_exception_code():
    exception = APIAuthException("Unexpected error")
    assert exception.detail == "Unexpected error"
    assert exception.status_code == 401


api_default = EasyAPI(urls_namespace="exception_default", easy_output=False)


@api_default.get("/list_exception")
async def list_exception_default(request):
    raise BaseAPIException(
        [
            "some error 1",
            "some error 2",
        ]
    )


client_default = testing.EasyTestClient(api_default)


async def test_exception_with_list_details_default():
    res = await client_default.get("list_exception")
    assert res.status_code == 500
    assert res.json() == {
        "detail": "[ErrorDetail(string='some error 1', code='error'), "
        "ErrorDetail(string='some error 2', code='error')]",
    }



import copy

import pytest
from django.contrib.auth import get_user_model

from easy.controller.base import CrudAPIController
from easy.testing import EasyTestClient

from .easy_app.auth import JWTAuthAsync, jwt_auth_async
from .easy_app.factories import UserFactory

User = get_user_model()


@pytest.fixture(autouse=True)
def media_storage(settings, tmpdir):
    settings.MEDIA_ROOT = tmpdir.strpath


@pytest.fixture
def user(db) -> User:
    return UserFactory()


@pytest.fixture
def easy_api_client(user) -> EasyTestClient:
    orig_func = copy.deepcopy(JWTAuthAsync.__call__)

    async def mock_func(self, request):
        setattr(request, "user", user)
        return True

    setattr(JWTAuthAsync, "__call__", mock_func)

    def create_client(
        api: CrudAPIController,
        is_staff: bool = False,
        is_superuser: bool = False,
    ):
        setattr(user, "is_staff", is_staff)
        setattr(user, "is_superuser", is_superuser)
        client = EasyTestClient(api, auth=jwt_auth_async)
        return client

    yield create_client
    setattr(JWTAuthAsync, "__call__", orig_func)



from typing import Any

from django.conf import settings as django_settings
from django.test.signals import setting_changed

# AUTO ADMIN API settings
# If not all
AUTO_ADMIN_ENABLED_ALL_APPS = getattr(
    django_settings, "AUTO_ADMIN_ENABLED_ALL_APPS", True
)
# Only generate for included apps
AUTO_ADMIN_INCLUDE_APPS = getattr(django_settings, "AUTO_ADMIN_INCLUDE_APPS", [])
# Exclude apps always got excluded
AUTO_ADMIN_EXCLUDE_APPS = getattr(django_settings, "AUTO_ADMIN_EXCLUDE_APPS", [])


def reload_settings(*args: Any, **kwargs: Any) -> None:  # pragma: no cover
    global settings

    setting, value = kwargs["setting"], kwargs["value"]
    globals()[setting] = value


setting_changed.connect(reload_settings)  # pragma: no cover



from typing import TYPE_CHECKING, Any

from django.http import HttpRequest
from ninja_extra import permissions

if TYPE_CHECKING:
    from ninja_extra.controllers.base import ControllerBase  # pragma: no cover


class BaseApiPermission(permissions.BasePermission):
    """
    Base permission class that all Permission Class should inherit from.
    This will call service.check_permission for extra check.
    """

    def has_permission(
        self, request: HttpRequest, controller: "ControllerBase"
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        has_perm: bool = True
        if hasattr(controller, "service"):
            has_perm = controller.service.check_permission(request, controller)  # type: ignore
        return has_perm

    def has_object_permission(
        self, request: HttpRequest, controller: "ControllerBase", obj: Any
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        has_perm: bool = True
        if hasattr(controller, "service"):
            has_perm = controller.service.check_object_permission(  # type: ignore
                request, controller, obj
            )
        return has_perm



from datetime import date

from ninja import Schema


class EventSchema(Schema):
    title: str
    start_date: date
    end_date: date

    class Config:
        orm_mode = True


class EventSchemaOut(Schema):
    id: int



import logging
from abc import ABC
from typing import Any

from asgiref.sync import sync_to_async
from django.contrib.auth import get_user_model
from django.http import HttpRequest
from ninja_jwt.authentication import JWTAuth

logger = logging.getLogger(__name__)
user_model = get_user_model()


class JWTAuthAsync(JWTAuth, ABC):
    async def authenticate(self, request: HttpRequest, token: str) -> Any:
        return await sync_to_async(super().authenticate)(request, token)


jwt_auth = JWTAuth()
jwt_auth_async = JWTAuthAsync()



import json
from typing import Any, Dict, Union

from django.http.response import JsonResponse

from easy.renderer.json import EasyJSONEncoder

ERRNO_SUCCESS = 0
SUCCESS_MESSAGE = "success"
UNKNOWN_ERROR_MSG = "system error"


class BaseApiResponse(JsonResponse):
    """
    Base for all API responses
    """

    def __init__(
        self,
        data: Union[Dict, str] = None,
        errno: int = None,
        message: str = None,
        **kwargs: Any
    ):
        if errno:
            message = message or UNKNOWN_ERROR_MSG
        else:
            message = SUCCESS_MESSAGE
            errno = ERRNO_SUCCESS

        _data: Union[Dict, str] = {
            "code": errno,
            "message": message,
            "data": data if data is not None else {},
        }

        super().__init__(data=_data, encoder=EasyJSONEncoder, **kwargs)

    @property
    def json_data(self) -> Any:
        """
        Get json data
        """
        return json.loads(self.content)

    def update_content(self, data: Dict) -> None:
        """
        Update content with new data
        """
        self.content = json.dumps(data)



from ninja_extra.operation import AsyncOperation

from easy import EasyAPI

api_admin_v1 = EasyAPI()
api_admin_v1.auto_create_admin_controllers()

path_names = []
controllers = []
controller_names = []

for path, rtr in api_admin_v1._routers:
    path_names.append(path)
    controllers.append(rtr)
    controller_names.append(str(rtr))
    for path_ops in rtr.path_operations.values():
        for op in path_ops.operations:
            assert isinstance(op, AsyncOperation)
            assert op.api is api_admin_v1


def test_auto_generate_admin_api():
    assert len(api_admin_v1._routers) == 5  # default + 3 models
    assert "/easy_app/category" in path_names
    assert "/easy_app/client" in path_names
    assert "/easy_app/event" in path_names
    assert "/easy_app/type" in path_names

    assert "CategoryAdminAPIController" in controller_names
    assert "EventAdminAPIController" in controller_names
    assert "ClientAdminAPIController" in controller_names
    assert "TypeAdminAPIController" in controller_names


async def test_auto_apis(transactional_db, easy_api_client):
    for controller_class in controllers:
        if not str(controller_class).endswith("ClientAdminAPIController"):
            continue
        client = easy_api_client(controller_class)
        response = await client.get("/")
        assert response.status_code == 200
        assert response.json()["data"] == []

        response = await client.delete("/20000")
        assert response.status_code == 200
        assert response.json()["code"] == 404


async def test_auto_generation_settings(settings):
    settings.AUTO_ADMIN_EXCLUDE_APPS = ["tests.easy_app"]
    api_admin_v2 = EasyAPI()
    api_admin_v2.auto_create_admin_controllers()
    assert len(api_admin_v2._routers) == 1

    settings.AUTO_ADMIN_ENABLED_ALL_APPS = False
    settings.AUTO_ADMIN_EXCLUDE_APPS = []
    settings.AUTO_ADMIN_INCLUDE_APPS = ["tests.none_existing_app"]
    api_admin_v3 = EasyAPI()
    api_admin_v3.auto_create_admin_controllers()
    assert len(api_admin_v3._routers) == 1



import logging
from importlib import import_module
from typing import Any, Callable, Optional, Sequence, Union

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.utils.module_loading import module_has_submodule
from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.parser import Parser
from ninja.renderers import BaseRenderer
from ninja.types import TCallable
from ninja_extra import NinjaExtraAPI

from easy.controller.admin_auto_api import create_admin_controller
from easy.domain.serializers import django_serializer
from easy.renderer.json import EasyJSONRenderer
from easy.response import BaseApiResponse

logger = logging.getLogger(__name__)


class EasyAPI(NinjaExtraAPI):
    """
    EasyAPI, extensions:
    -Add 2 init params
        Easy_extra: bool = True,
            Can serialize queryset or model, and support pagination
        Easy_output: bool = True,
            If True, will be encapsulated in BaseAPIResponse
    -renderer, default to EasyJSONRenderer
    -Auto generate AdminAPIs, it will read the following settings:
        AUTO_ADMIN_ENABLED_ALL_APPS
        AUTO_ADMIN_EXCLUDE_APPS
        AUTO_ADMIN_INCLUDE_APPS
    """

    def __init__(
        self,
        *,
        title: str = "Easy API",
        version: str = "1.0.0",
        description: str = "",
        openapi_url: Optional[str] = "/openapi.json",
        docs_url: Optional[str] = "/docs",
        docs_decorator: Optional[Callable[[TCallable], TCallable]] = None,
        urls_namespace: Optional[str] = None,
        csrf: bool = False,
        auth: Union[
            Sequence[Callable[..., Any]], Callable[..., Any], NOT_SET_TYPE, None
        ] = NOT_SET,
        renderer: Optional[BaseRenderer] = EasyJSONRenderer(),
        parser: Optional[Parser] = None,
        app_name: str = "ninja",
        easy_extra: bool = True,
        easy_output: bool = True,
    ) -> None:
        super(NinjaExtraAPI, self).__init__(
            title=title,
            version=version,
            description=description,
            openapi_url=openapi_url,
            docs_url=docs_url,
            docs_decorator=docs_decorator,
            urls_namespace=urls_namespace,
            csrf=csrf,
            auth=auth,
            renderer=renderer,
            parser=parser,
        )
        self.docs_decorator = docs_decorator
        self.app_name = app_name
        self.easy_extra = easy_extra
        self.easy_output = easy_output

    def auto_create_admin_controllers(self, version: str = None) -> None:
        for app_module in self.get_installed_apps():
            # If not all
            if not settings.AUTO_ADMIN_ENABLED_ALL_APPS:  # type:ignore
                # Only generate for this included apps
                if settings.AUTO_ADMIN_INCLUDE_APPS is not None:  # type:ignore
                    if (
                        app_module.name
                        not in settings.AUTO_ADMIN_INCLUDE_APPS  # type:ignore
                    ):
                        continue

            # Exclude list
            if app_module.name in settings.AUTO_ADMIN_EXCLUDE_APPS:  # type:ignore
                continue

            try:
                app_module_ = import_module(app_module.name)
                final = []
                if module_has_submodule(app_module_, "models"):
                    # Auto generate AdminAPI
                    for model in app_module.get_models():
                        final.append(
                            create_admin_controller(
                                model, app_module.name.split(".")[1]
                            )
                        )
                self.register_controllers(*final)
            except ImportError as ex:  # pragma: no cover
                raise ex

    @staticmethod
    def get_installed_apps() -> list:
        from django.apps import apps

        return [
            v
            for k, v in apps.app_configs.items()
            if not v.name.startswith("django.") and (not v.name == "easy.api")
        ]

    def create_response(
        self,
        request: HttpRequest,
        data: Any,
        *,
        status: int = None,
        temporal_response: HttpResponse = None,
    ) -> HttpResponse:
        if self.easy_extra:
            try:
                data = django_serializer.serialize_data(data)
            except Exception as e:  # pragma: no cover
                logger.error(f"Creat Response Error - {e}", exc_info=True)
                return BaseApiResponse(str(e), errno=500)

        if self.easy_output:
            if temporal_response:
                status = temporal_response.status_code
            assert status

            _temp = BaseApiResponse(
                data, status=status, content_type=self.get_content_type()
            )

            if temporal_response:
                response = temporal_response
                response.content = _temp.content
            else:
                response = _temp

        else:
            response = super().create_response(
                request,
                data,
                status=status,
                temporal_response=temporal_response,
            )
        return response

    def create_temporal_response(self, request: HttpRequest) -> HttpResponse:
        if self.easy_output:
            return BaseApiResponse("", content_type=self.get_content_type())
        else:
            return super().create_temporal_response(request)



import logging
from typing import Optional, Type, Union

from django.db import models
from ninja_extra import ControllerBase, api_controller
from ninja_extra.permissions import BasePermission

from easy.controller.base import CrudAPIController
from easy.controller.meta_conf import (
    GENERATE_CRUD_ATTR_DEFAULT,
    MODEL_FIELDS_ATTR_DEFAULT,
    MODEL_JOIN_ATTR_DEFAULT,
    MODEL_RECURSIVE_ATTR_DEFAULT,
    SENSITIVE_FIELDS_ATTR_DEFAULT,
)
from easy.permissions import AdminSitePermission, BaseApiPermission

logger = logging.getLogger(__name__)


def create_api_controller(
    model: models.Model,
    app_name: str,
    permission_class: Type[BasePermission] = BaseApiPermission,
    controller_name_prefix: Optional[str] = None,
) -> Union[Type[ControllerBase], Type]:
    """Create APIController class dynamically, with specified permission class"""
    model_name = model.__name__  # type:ignore
    Meta = type(
        "Meta",
        (object,),
        {
            "model": model,
            "generate_crud": GENERATE_CRUD_ATTR_DEFAULT,
            "model_fields": MODEL_FIELDS_ATTR_DEFAULT,
            "model_recursive": MODEL_RECURSIVE_ATTR_DEFAULT,
            "model_join": MODEL_JOIN_ATTR_DEFAULT,
            "sensitive_fields": SENSITIVE_FIELDS_ATTR_DEFAULT,
        },
    )

    class_name = f"{model_name}{controller_name_prefix}APIController"

    auto_cls = type.__new__(
        type,
        class_name,
        (CrudAPIController,),
        {
            "Meta": Meta,
        },
    )

    return api_controller(
        f"/{app_name}/{model_name.lower()}",
        tags=[f"{model_name} {controller_name_prefix}API"],
        permissions=[permission_class],
    )(auto_cls)


def create_admin_controller(
    model: models.Model, app_name: str
) -> Union[Type[ControllerBase], Type]:
    """Create AdminAPI class dynamically, permission class set to AdminSitePermission"""
    return create_api_controller(
        model=model,
        app_name=app_name,
        permission_class=AdminSitePermission,
        controller_name_prefix="Admin",
    )



from easy.conf.settings import (
    AUTO_ADMIN_ENABLED_ALL_APPS,
    AUTO_ADMIN_EXCLUDE_APPS,
    AUTO_ADMIN_INCLUDE_APPS,
)


def test_change_django_settings(settings):
    assert settings.AUTO_ADMIN_ENABLED_ALL_APPS == AUTO_ADMIN_ENABLED_ALL_APPS
    assert settings.AUTO_ADMIN_INCLUDE_APPS == AUTO_ADMIN_INCLUDE_APPS
    assert settings.AUTO_ADMIN_EXCLUDE_APPS == AUTO_ADMIN_EXCLUDE_APPS



import logging
from typing import Type

from django.db import models

from easy.services.crud import CrudService
from easy.services.permission import PermissionService

logger = logging.getLogger(__name__)


class BaseService(CrudService, PermissionService):
    def __init__(
        self,
        model: Type[models.Model],
    ):
        self.model = model
        super().__init__(model=self.model)



import json

import pytest

from easy.response import BaseApiResponse


def test_base_api_result_base():

    assert BaseApiResponse("").json_data["data"] == ""

    assert BaseApiResponse("1").json_data["data"] == "1"

    assert BaseApiResponse("0").json_data["data"] == "0"
    assert BaseApiResponse().json_data["data"] == {}
    assert BaseApiResponse([]).json_data["data"] == []
    assert BaseApiResponse(True).json_data["data"] is True
    assert BaseApiResponse(False).json_data["data"] is False
    assert BaseApiResponse([1, 2, 3]).json_data["data"] == [1, 2, 3]


def test_base_api_result_dict():

    assert BaseApiResponse({"a": 1, "b": 2}).json_data["data"] == {
        "a": 1,
        "b": 2,
    }

    assert (BaseApiResponse({"code": 2, "im": 14})).json_data["data"]["im"] == 14
    assert (BaseApiResponse({"code": 2, "im": 14})).json_data["data"]["code"] == 2


def test_base_api_result_message():
    assert (
        BaseApiResponse(errno=-1, message="error test").json_data["message"]
        == "error test"
    )
    assert BaseApiResponse().json_data["message"]


def test_base_api_edit():
    orig_resp = BaseApiResponse(
        {"item_id": 2, "im": 14},
        errno=0,
    )

    with pytest.raises(KeyError):
        orig_resp.json_data["detail"]

    data = orig_resp.json_data

    data["detail"] = "Edited!!!"
    orig_resp.content = json.dumps(data)

    assert orig_resp.json_data["detail"] == "Edited!!!"

    data = orig_resp.json_data
    assert data["code"] == 0

    data["code"] = 401
    orig_resp.update_content(data)
    assert orig_resp.json_data["code"] == 401
    assert orig_resp.json_data["data"]["im"] == 14

    data["data"]["im"] = 8888
    orig_resp.update_content(data)
    assert orig_resp.json_data["data"]["im"] == 8888



from django.contrib import admin
from django.urls import path
from ninja_extra import NinjaExtraAPI

from .controllers import (
    AutoGenCrudAPIController,
    EasyCrudAPIController,
    PermissionAPIController,
)

api = NinjaExtraAPI()
api.register_controllers(EasyCrudAPIController)
api.register_controllers(PermissionAPIController)
api.register_controllers(AutoGenCrudAPIController)


urlpatterns = [
    path("admin/", admin.site.urls),
    path("api/", api.urls),
]



from typing import Any, List, Optional, Type, Union

from django.db import models

META_ATTRIBUTE_NAME: str = "__EASY_API_META__"

GENERATE_CRUD_ATTR: str = "generate_crud"
GENERATE_CRUD_ATTR_DEFAULT = True

MODEL_EXCLUDE_ATTR: str = "model_exclude"
MODEL_EXCLUDE_ATTR_DEFAULT: List = []

MODEL_FIELDS_ATTR: str = "model_fields"
MODEL_FIELDS_ATTR_DEFAULT: str = "__all__"

MODEL_RECURSIVE_ATTR: str = "model_recursive"
MODEL_RECURSIVE_ATTR_DEFAULT: bool = False

MODEL_JOIN_ATTR: str = "model_join"
MODEL_JOIN_ATTR_DEFAULT: bool = False

SENSITIVE_FIELDS_ATTR: str = "sensitive_fields"
SENSITIVE_FIELDS_ATTR_DEFAULT: List = ["password", "token"]


class ModelOptions:
    def __init__(self, options: object = None):
        """
        Configuration reader
        """
        self.model: Optional[Type[models.Model]] = getattr(options, "model", None)
        self.generate_crud: Optional[Union[bool]] = getattr(
            options, GENERATE_CRUD_ATTR, GENERATE_CRUD_ATTR_DEFAULT
        )
        self.model_exclude: Union[Union[str], List[Any]] = getattr(
            options, MODEL_EXCLUDE_ATTR, MODEL_EXCLUDE_ATTR_DEFAULT
        )
        self.model_fields: Union[Union[str], List[Any]] = getattr(
            options, MODEL_FIELDS_ATTR, MODEL_FIELDS_ATTR_DEFAULT
        )
        self.model_join: Optional[Union[bool]] = getattr(
            options, MODEL_JOIN_ATTR, MODEL_JOIN_ATTR_DEFAULT
        )
        self.model_recursive: Optional[Union[bool]] = getattr(
            options, MODEL_RECURSIVE_ATTR, MODEL_RECURSIVE_ATTR_DEFAULT
        )
        self.sensitive_fields: Optional[Union[str, List[str]]] = getattr(
            options, SENSITIVE_FIELDS_ATTR, list(SENSITIVE_FIELDS_ATTR_DEFAULT)
        )

    @classmethod
    def get_model_options(cls, klass: Type) -> Any:
        return ModelOptions(getattr(klass, "Meta", None))

    @classmethod
    def set_model_meta(
        cls, model: Type[models.Model], model_opts: "ModelOptions"
    ) -> None:
        setattr(
            model,
            META_ATTRIBUTE_NAME,
            {
                GENERATE_CRUD_ATTR: model_opts.generate_crud,
                MODEL_EXCLUDE_ATTR: model_opts.model_exclude,
                MODEL_FIELDS_ATTR: model_opts.model_fields,
                MODEL_RECURSIVE_ATTR: model_opts.model_recursive,
                MODEL_JOIN_ATTR: model_opts.model_join,
                SENSITIVE_FIELDS_ATTR: model_opts.sensitive_fields,
            },
        )


class ModelMetaConfig(object):
    @staticmethod
    def get_configuration(obj: models.Model, _name: str, default: Any = None) -> Any:
        _value = default if default else None
        if hasattr(obj, META_ATTRIBUTE_NAME):
            _value = getattr(obj, META_ATTRIBUTE_NAME).get(_name, _value)
        return _value

    def get_model_recursive(self, obj: models.Model) -> bool:
        model_recursive: bool = self.get_configuration(
            obj, MODEL_RECURSIVE_ATTR, default=MODEL_RECURSIVE_ATTR_DEFAULT
        )
        return model_recursive

    def get_model_join(self, obj: models.Model) -> bool:
        model_join: bool = self.get_configuration(
            obj, MODEL_JOIN_ATTR, default=MODEL_JOIN_ATTR_DEFAULT
        )
        return model_join

    def get_model_fields_list(self, obj: models.Model) -> List[Any]:
        model_fields: List = self.get_configuration(
            obj, MODEL_FIELDS_ATTR, default=MODEL_FIELDS_ATTR_DEFAULT
        )
        return model_fields

    def get_model_exclude_list(self, obj: models.Model) -> List[Any]:
        exclude_list: List = self.get_configuration(
            obj, MODEL_EXCLUDE_ATTR, default=MODEL_EXCLUDE_ATTR_DEFAULT
        )
        return exclude_list

    def get_sensitive_list(self, obj: models.Model) -> List[Any]:
        sensitive_list: List = self.get_configuration(
            obj, SENSITIVE_FIELDS_ATTR, default=SENSITIVE_FIELDS_ATTR_DEFAULT
        )
        return sensitive_list

    def get_final_excluded_list(self, obj: models.Model) -> List[Any]:
        total_excluded_list = []
        sensitive_list: List = list(SENSITIVE_FIELDS_ATTR_DEFAULT)
        excluded_list = []

        sensitive_fields = self.get_sensitive_list(obj)
        if sensitive_fields:
            sensitive_list.extend(sensitive_fields)
        sensitive_list = list(set(sensitive_list))

        excluded_fields = self.get_model_exclude_list(obj)
        if excluded_fields:
            excluded_list.extend(excluded_fields)
        excluded_list = list(set(excluded_list))

        total_excluded_list.extend(sensitive_list)
        total_excluded_list.extend(excluded_list)
        return list(set(total_excluded_list))

    def show_field(self, obj: models.Model, field_name: str) -> bool:
        model_exclude_list = self.get_model_exclude_list(obj)
        if model_exclude_list:
            if field_name in self.get_final_excluded_list(obj):
                return False
        else:
            if field_name in self.get_final_excluded_list(obj):
                return False
            model_fields_list = self.get_model_fields_list(obj)
            if model_fields_list != MODEL_FIELDS_ATTR_DEFAULT:
                if field_name not in model_fields_list:
                    return False
        return True



from ninja_extra.permissions import (
    AllowAny,
    IsAdminUser,
    IsAuthenticated,
    IsAuthenticatedOrReadOnly,
)

from .adminsite import AdminSitePermission
from .base import BaseApiPermission
from .superuser import IsSuperUser

__all__ = [
    "AllowAny",
    "AdminSitePermission",
    "BaseApiPermission",
    "IsAuthenticated",
    "IsAuthenticatedOrReadOnly",
    "IsAdminUser",
    "IsSuperUser",
]



import functools
import types


def copy_func(f):  # type: ignore
    n = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    n = functools.update_wrapper(n, f)
    n.__kwdefaults__ = f.__kwdefaults__
    return n



from ninja_extra import status
from ninja_extra.exceptions import APIException


class BaseAPIException(APIException):
    """
    API Exception
    """

    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = (
        "There is an unexpected error, please try again later, if the problem "
        "persists, please contact customer support team for further support."
    )


class APIAuthException(BaseAPIException):
    """
    API Auth Exception
    """

    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "Unauthorized"



from easy.domain import BaseDomain

from .models import Event


class EventDomain(BaseDomain):
    pass


class EventBiz(EventDomain):
    def __init__(self, model=Event):
        self.model = model
        super(EventBiz, self).__init__(self.model)



from typing import List

from asgiref.sync import sync_to_async
from ninja_extra import api_controller, http_get, paginate

from easy.controller.base import CrudAPIController
from easy.permissions import (
    AdminSitePermission,
    BaseApiPermission,
    IsAdminUser,
    IsAuthenticated,
    IsSuperUser,
)
from easy.response import BaseApiResponse

from .models import Client, Event
from .schema import EventSchema
from .services import EventService


@api_controller("unittest", permissions=[BaseApiPermission])
class AutoGenCrudAPIController(CrudAPIController):
    """
    For unit testings of the following auto generated APIs:
        get/create/patch/delete/filter/filter_exclude
    """

    def __init__(self, service: EventService):
        super().__init__(service)

    class Meta:
        model = Event
        model_join = True


@api_controller("unittest", permissions=[BaseApiPermission])
class RecursiveAPIController(CrudAPIController):
    """
    For unit testings of no recursive configuration
    """

    def __init__(self, service: EventService):
        super().__init__(service)

    class Meta:
        model = Event
        model_fields = "__all__"
        model_join = True
        model_recursive = True


@api_controller("unittest", permissions=[BaseApiPermission])
class InheritedRecursiveAPIController(AutoGenCrudAPIController):
    """
    For unit testings of no recursive configuration
    """

    def __init__(self, service: EventService):
        super().__init__(service)

    class Meta:
        model = Event
        model_fields = "__all__"
        model_join = True
        model_recursive = True


@api_controller("unittest", permissions=[BaseApiPermission])
class AutoGenCrudNoJoinAPIController(CrudAPIController):
    """
    For unit testings of mo model_join and sensitive_fields configuration
    """

    def __init__(self, service: EventService):
        super().__init__(service)

    class Meta:
        model = Event
        model_fields = "__all__"
        model_join = False
        model_recursive = True
        sensitive_fields = ["password", "sensitive_info"]


@api_controller("unittest", permissions=[BaseApiPermission])
class AutoGenCrudSomeFieldsAPIController(CrudAPIController):
    """
    For unit testings of the no-m2m-fields model
    """

    class Meta:
        model = Client
        model_fields = [
            "key",
            "name",
        ]


@api_controller("unittest")
class EasyCrudAPIController(CrudAPIController):
    """
    For unit testings of demo APIs
    """

    def __init__(self, service: EventService):
        super().__init__(service)

    class Meta:
        model = Event
        model_exclude = [
            "category",
        ]

    @http_get("/base_response/")
    async def generate_base_response(self, request):
        return BaseApiResponse({"data": "This is a BaseApiResponse."})

    @http_get("/qs_paginated/", auth=None)
    @paginate
    async def qs_paginated(self, request):
        return await self.service.get_event_objs_demo()

    @http_get("/qs_list/", response=List[EventSchema])
    async def get_objs_list_with_filter_exclude(self, request):
        return await sync_to_async(list)(
            await self.service.filter_exclude_objs(
                title__endswith="qs_list",
            )
        )

    @http_get(
        "/qs/",
    )
    async def list_events(self):
        qs = await sync_to_async(self.model.objects.all)()
        await sync_to_async(list)(qs)
        if qs:
            return qs
        return BaseApiResponse()


@api_controller("unittest")
class PermissionAPIController(CrudAPIController):
    """
    For unit testings of permissions class
    """

    def __init__(self, service: EventService):
        super().__init__(service)
        self.service = service

    class Meta:
        model = Event

    @http_get("/must_be_authenticated/", permissions=[IsAuthenticated])
    async def must_be_authenticated(self, word: str):
        return await self.service.get_identity_demo(word)

    @http_get("/must_be_admin_user/", permissions=[IsAdminUser])
    async def must_be_admin_user(self, word: str):
        return await self.service.get_identity_demo(word)

    @http_get("/must_be_super_user/", permissions=[IsSuperUser])
    async def must_be_super_user(self, word: str):
        return await self.service.get_identity_demo(word)

    @http_get("/test_perm_only_super/", permissions=[BaseApiPermission])
    async def test_perm_only_super(self, request):
        event_id = await self.service.add_obj(title="test_event_title")
        # return await self.service.get_obj(id=note.id)
        return await sync_to_async(self.get_object_or_none)(Event, id=event_id)

    @http_get("/test_perm/", permissions=[BaseApiPermission])
    async def test_perm(self, request, word: str):
        return await self.service.get_identity_demo(word)

    @http_get("/test_perm_admin_site/", permissions=[AdminSitePermission])
    async def test_perm_admin_site(self, request, word: str):
        return await self.service.get_identity_demo(word)


@api_controller("unittest", permissions=[AdminSitePermission])
class AdminSitePermissionAPIController(CrudAPIController):
    def __init__(self, service: EventService):
        super().__init__(service)
        self.service = service

    class Meta:
        model = Event


@api_controller("unittest", permissions=[AdminSitePermission])
class NoCrudAPIController(CrudAPIController):
    def __init__(self, service: EventService):
        super().__init__(service)
        self.service = service

    class Meta:
        model = Event
        generate_crud = False



from typing import TYPE_CHECKING

from django.http import HttpRequest

from .base import BaseApiPermission

if TYPE_CHECKING:
    from ninja_extra.controllers.base import ControllerBase  # pragma: no cover


class IsSuperUser(BaseApiPermission):
    """
    Allows access only to super user.
    """

    def has_permission(
        self, request: HttpRequest, controller: "ControllerBase"
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        user = request.user or request.auth  # type: ignore
        return bool(user and user.is_authenticated and user.is_superuser)  # type: ignore



import logging
from typing import Any, Type

from asgiref.sync import sync_to_async
from django.db import models

from easy.domain.orm import CrudModel

logger = logging.getLogger(__name__)


class CrudService(CrudModel):
    def __init__(self, model: Type[models.Model]):
        super().__init__(model)
        self.model = model

    async def get_obj(self, id: int) -> Any:
        return await sync_to_async(self._crud_get_obj)(id)

    async def get_objs(self, **filters: Any) -> Any:
        return await sync_to_async(self._crud_get_objs_all)(**filters)

    async def patch_obj(self, id: int, payload: Any) -> Any:
        return await sync_to_async(self._crud_update_obj)(id, payload)

    async def del_obj(self, id: int) -> Any:
        return await sync_to_async(self._crud_del_obj)(id)

    async def add_obj(self, **payload: Any) -> Any:
        return await sync_to_async(self._crud_add_obj)(**payload)

    async def filter_objs(self, **payload: Any) -> Any:
        return await sync_to_async(self._crud_filter)(**payload)  # pragma: no cover

    async def filter_exclude_objs(self, **payload: Any) -> Any:
        return await sync_to_async(self._crud_filter_exclude)(**payload)

    # async def bulk_create_objs(self):
    #     ...
    #
    # async def recover_obj(self):
    #     ...



from typing import TYPE_CHECKING

from django.http import HttpRequest
from ninja_extra.permissions import IsAdminUser

if TYPE_CHECKING:
    from ninja_extra.controllers.base import ControllerBase  # pragma: no cover


class AdminSitePermission(IsAdminUser):
    """
    Allows delete only to super user, and change to only staff/super users.
    """

    def has_permission(
        self, request: HttpRequest, controller: "ControllerBase"
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        user = request.user or request.auth  # type: ignore
        has_perm: bool = True
        if request.method == "DELETE":
            has_perm = bool(user.is_superuser)  # type: ignore
        if request.method in ("PUT", "PATCH", "POST"):
            has_perm = bool(user.is_staff or user.is_superuser)  # type: ignore
        return has_perm and super().has_permission(request, controller)


#
# class AdminSitePermissionV2(AdminSitePermission):
#     def has_permission(
#         self, request: HttpRequest, controller: "ControllerBase"
#     ) -> bool:
#         """
#         Return `True` if permission is granted, `False` otherwise.
#         """
#         user = request.user
#         has_perm = True
#         if hasattr(controller, "model"):
#             model = controller.model
#             app = model._meta.app_label
#             has_perm = user.has_perm(f"{app}.view_{model.__name__}")
#             if request.method in ("PUT",):
#                 has_perm = user.has_perm(f"{app}.add_{model.__name__}")
#             if request.method in ("PATCH", "POST"):
#                 has_perm = user.has_perm(f"{app}.change_{model.__name__}")
#             if request.method in ("DELETE",):
#                 has_perm = user.has_perm(f"{app}.delete_{model.__name__}")
#         if user.is_superuser:
#             has_perm = True
#         return bool(user and user.is_authenticated and user.is_active and has_perm)
#
#     async def has_permission(
#         self, request: HttpRequest, controller: "ControllerBase"
#     ) -> bool:
#         """
#         Return `True` if permission is granted, `False` otherwise.
#         """
#         user = request.user
#         has_perm = False
#         if hasattr(controller, "model"):
#             model = controller.model
#             app = model._meta.app_label
#             has_perm = await sync_to_async(user.has_perm)(
#                 f"{app}.view_{model.__name__}"
#             )
#             if request.method in ("PUT",):
#                 has_perm = await sync_to_async(user.has_perm)(
#                     f"{app}.add_{model.__name__}"
#                 )
#             if request.method in ("PATCH", "POST"):
#                 has_perm = await sync_to_async(user.has_perm)(
#                     f"{app}.change_{model.__name__}"
#                 )
#             if request.method in ("DELETE",):
#                 has_perm = await sync_to_async(user.has_perm)(
#                     f"{app}.delete_{model.__name__}"
#                 )
#         if user.is_superuser:
#             has_perm = True
#         return bool(user and user.is_authenticated and user.is_active and has_perm)



"""
With these settings, tests run faster.
"""

# Admin API auto-generation settings
AUTO_ADMIN_ENABLED_ALL_APPS = True
AUTO_ADMIN_INCLUDE_APPS = []
AUTO_ADMIN_EXCLUDE_APPS = []

# Django Settings
INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "ninja_extra",
    "tests.easy_app",
    "easy",
)

MIDDLEWARE = (
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

USE_I18N = True
USE_TZ = True
LANGUAGE_CODE = "en-us"

STATIC_URL = "/static/"
ROOT_URLCONF = "tests.easy_app.urls"

AUTHENTICATION_BACKENDS = ("django.contrib.auth.backends.ModelBackend",)

# DB
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
        # "TEST": {
        #     # this gets you in-memory sqlite for faster testing
        #     "ENGINE": "django.db.backends.sqlite3",
        # },
    }
}

# https://docs.djangoproject.com/en/dev/ref/settings/#test-runner
TEST_RUNNER = "django.test.runner.DiscoverRunner"

# CACHES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#caches
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "",
    }
}

# PASSWORDS
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#password-hashers
PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]

# TEMPLATES
# ------------------------------------------------------------------------------
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# EMAIL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#email-backend
EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"


# Security
ALLOWED_HOSTS = ["*"]
SECRET_KEY = "not very secret in tests"

# Debug
DEBUG_PROPAGATE_EXCEPTIONS = True



from datetime import datetime, timedelta

import django
import pytest
from asgiref.sync import sync_to_async

from .easy_app.controllers import (
    AdminSitePermissionAPIController,
    AutoGenCrudAPIController,
    PermissionAPIController,
)
from .easy_app.models import Client, Event
from .test_async_other_apis import dummy_data


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
@pytest.mark.django_db
class TestPermissionController:
    async def test_demo(self, easy_api_client):
        client = easy_api_client(PermissionAPIController)

        response = await client.get(
            "/must_be_authenticated/?word=authenticated",
            content_type="application/json",
        )
        assert response.status_code == 200
        assert response.json().get("data")["says"] == "authenticated"

        client = easy_api_client(PermissionAPIController)
        response = await client.get(
            "/must_be_admin_user/?word=admin",
        )
        assert response.status_code == 403
        with pytest.raises(KeyError):
            assert response.json().get("data")["says"] == "admin"

        client = easy_api_client(PermissionAPIController, is_staff=True)
        response = await client.get(
            "/must_be_admin_user/?word=admin",
        )
        assert response.status_code == 200
        assert response.json().get("data")["says"] == "admin"

        client = easy_api_client(PermissionAPIController)
        response = await client.get(
            "/must_be_super_user/?word=superuser",
        )
        assert response.status_code == 403
        with pytest.raises(KeyError):
            assert response.json().get("data")["says"] == "superuser"

        client = easy_api_client(PermissionAPIController, is_superuser=True)
        response = await client.get(
            "/must_be_super_user/?word=superuser",
        )
        assert response.status_code == 200
        assert response.json().get("data")["says"] == "superuser"

    async def test_perm(self, transactional_db, easy_api_client):
        client = easy_api_client(PermissionAPIController)
        response = await client.get("/test_perm/", query=dict(word="normal"))
        assert response.status_code == 200
        assert response.json().get("data")["says"] == "normal"
        client = easy_api_client(PermissionAPIController, is_staff=True)
        response = await client.get("/test_perm/", query=dict(word="staff"))
        assert response.status_code == 200
        assert response.json().get("data")["says"] == "staff"

    async def test_perm_only_super(self, transactional_db, easy_api_client):
        client = easy_api_client(PermissionAPIController)
        response = await client.get("/test_perm_only_super/")
        assert response.status_code == 403
        assert response.json().get("data") == {
            "detail": "You do not have permission to perform this action."
        }

        client = easy_api_client(PermissionAPIController)
        response = await client.get("/test_perm_only_super/")
        assert response.status_code == 403
        assert response.json().get("data") == {
            "detail": "You do not have permission to perform this action."
        }

        client = easy_api_client(PermissionAPIController, is_superuser=True)
        response = await client.get("/test_perm_only_super/")
        assert response.status_code == 200
        assert response.json().get("data")["title"] == "test_event_title"

    async def test_perm_admin_site(self, transactional_db, easy_api_client):
        # None-admin users
        client = easy_api_client(PermissionAPIController)
        response = await client.get(
            "/test_perm_admin_site/", query=dict(word="non-admin")
        )
        assert response.status_code == 403
        assert response.json().get("data") == {
            "detail": "You do not have permission to perform this action."
        }

        # Staff users
        client = easy_api_client(PermissionAPIController, is_staff=True)
        response = await client.get("/test_perm_admin_site/", query=dict(word="staff"))
        assert response.status_code == 200
        assert response.json()["data"]["says"] == "staff"

    async def test_perm_auto_apis_delete(self, transactional_db, easy_api_client):
        client = easy_api_client(AdminSitePermissionAPIController)
        # Test delete
        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get")
        event = await sync_to_async(Event.objects.create)(**object_data)
        response = await client.get(
            f"/{event.id}",
        )
        assert response.status_code == 403

        response = await client.delete(
            f"/{event.id}",
        )
        assert response.status_code == 403
        assert response.json().get("data") == {
            "detail": "You do not have permission to perform this action."
        }

        # Super users
        client = easy_api_client(AutoGenCrudAPIController, is_superuser=True)
        response = await client.delete(
            f"/{event.id}",
        )

        assert response.status_code == 200

        response = await client.get(
            f"/{event.id}",
        )
        assert response.status_code == 200
        assert response.json().get("code") == 404

    async def test_perm_auto_apis_patch(self, transactional_db, easy_api_client):
        client = easy_api_client(AdminSitePermissionAPIController)

        object_data = dummy_data.copy()
        event = await sync_to_async(Event.objects.create)(**object_data)

        response = await client.get(
            f"/{event.id}",
        )
        assert response.status_code == 403
        assert response.json().get("data") == {
            "detail": "You do not have permission to perform this action."
        }

        # Staff users
        client = easy_api_client(AutoGenCrudAPIController, is_staff=True)
        response = await client.get(
            f"/{event.id}",
        )
        assert response.json().get("data")["title"] == f"{object_data['title']}"

        client_g = await sync_to_async(Client.objects.create)(
            name="Client G for Unit Testings", key="G"
        )

        client_h = await sync_to_async(Client.objects.create)(
            name="Client H for Unit Testings", key="H"
        )

        new_data = dict(
            id=event.id,
            title=f"{object_data['title']}_patch",
            start_date=str((datetime.now() + timedelta(days=10)).date()),
            end_date=str((datetime.now() + timedelta(days=20)).date()),
            owner=[client_g.id, client_h.id],
        )

        client = easy_api_client(AdminSitePermissionAPIController)
        response = await client.patch(
            f"/{event.id}", json=new_data, content_type="application/json"
        )

        assert response.status_code == 403
        assert response.json().get("data") == {
            "detail": "You do not have permission to perform this action."
        }

        # Super users
        client = easy_api_client(AutoGenCrudAPIController, is_superuser=True)
        response = await client.patch(
            f"/{event.id}", json=new_data, content_type="application/json"
        )
        assert response.json().get("data")

        response = await client.get(
            f"/{event.id}",
        )
        assert response.status_code == 200
        assert response.json().get("data")["title"] == "AsyncAPIEvent_patch"
        assert response.json().get("data")["start_date"] == str(
            (datetime.now() + timedelta(days=10)).date()
        )
        assert response.json().get("data")["end_date"] == str(
            (datetime.now() + timedelta(days=20)).date()
        )



from django.db import models


class TestBaseModel(models.Model):
    class Meta:
        abstract = True


class Category(TestBaseModel):
    title = models.CharField(max_length=100)
    status = models.PositiveSmallIntegerField(default=1, null=True)


class Client(TestBaseModel):
    key = models.CharField(max_length=20, unique=True)
    name = models.CharField(max_length=50, null=True)
    category = models.ForeignKey(Category, on_delete=models.CASCADE, null=True)
    password = models.CharField(max_length=30, null=True)


class Type(TestBaseModel):
    name = models.CharField(max_length=50, null=True)
    status = models.PositiveSmallIntegerField(default=1, null=True)


class Event(TestBaseModel):
    title = models.CharField(max_length=100)
    category = models.OneToOneField(
        Category, null=True, blank=True, on_delete=models.SET_NULL
    )
    start_date = models.DateField(
        null=True,
    )
    end_date = models.DateField(
        null=True,
    )
    photo = models.ImageField(upload_to="client/photo", null=True)

    owner = models.ManyToManyField(to=Client, related_name="events", null=True)

    lead_owner = models.ManyToManyField(to=Client, related_name="lead_owner", null=True)

    type = models.ForeignKey(Type, on_delete=models.CASCADE, null=True)

    sensitive_info = models.CharField(max_length=100, null=True)

    def __str__(self):
        return self.title



import logging
from typing import Any, Dict, List, Tuple

from django.db import models

from easy.controller.meta_conf import ModelMetaConfig

logger = logging.getLogger(__name__)


class DjangoSerializer(ModelMetaConfig):
    @staticmethod
    def is_model_instance(data: Any) -> bool:
        return isinstance(data, models.Model)

    @staticmethod
    def is_queryset(data: Any) -> bool:
        return isinstance(data, models.query.QuerySet)

    @staticmethod
    def is_one_relationship(data: Any) -> bool:
        return isinstance(data, models.ForeignKey) or isinstance(
            data, models.OneToOneRel
        )

    @staticmethod
    def is_many_relationship(data: Any) -> bool:
        return (
            isinstance(data, models.ManyToManyRel)
            or isinstance(data, models.ManyToManyField)
            or isinstance(data, models.ManyToOneRel)
        )

    @staticmethod
    def is_paginated(data: Any) -> bool:
        return isinstance(data, dict) and isinstance(
            data.get("items", None), models.query.QuerySet
        )

    def serialize_model_instance(
        self, obj: models.Model, referrers: Any = tuple()
    ) -> Dict[Any, Any]:
        """Serializes Django model instance to dictionary"""
        out = {}
        for field in obj._meta.get_fields():
            if self.show_field(obj, field.name):
                if self.is_one_relationship(field):
                    out.update(
                        self.serialize_foreign_key(obj, field, referrers + (obj,))
                    )

                elif self.is_many_relationship(field):
                    out.update(
                        self.serialize_many_relationship(obj, referrers + (obj,))
                    )

                else:
                    out.update(self.serialize_value_field(obj, field))
        return out

    def serialize_queryset(
        self, data: models.query.QuerySet, referrers: Tuple[Any, ...] = tuple()
    ) -> List[Dict[Any, Any]]:
        """Serializes Django Queryset to dictionary"""
        return [self.serialize_model_instance(obj, referrers) for obj in data]

    def serialize_foreign_key(
        self, obj: models.Model, field: Any, referrers: Any = tuple()
    ) -> Dict[Any, Any]:
        """Serializes foreign key field of Django model instance"""
        try:
            if not hasattr(obj, field.name):
                return {field.name: None}  # pragma: no cover
            related_instance = getattr(obj, field.name)
            if related_instance is None:
                return {field.name: None}
            if related_instance in referrers:
                return {}  # pragma: no cover
            field_value = getattr(related_instance, "pk")
        except Exception as exc:  # pragma: no cover
            logger.error(f"serialize_foreign_key error - {obj}", exc_info=exc)
            return {field.name: None}

        if self.get_model_recursive(obj):
            return {
                field.name: self.serialize_model_instance(related_instance, referrers)
            }
        return {field.name: field_value}

    def serialize_many_relationship(
        self, obj: models.Model, referrers: Any = tuple()
    ) -> Dict[Any, Any]:
        """
        Serializes many relationship (ManyToMany, ManyToOne) of Django model instance
        """
        if not hasattr(obj, "_prefetched_objects_cache"):
            return {}
        out = {}
        try:
            for k, v in obj._prefetched_objects_cache.items():  # type: ignore
                field_name = k if hasattr(obj, k) else k + "_set"
                if v:
                    if self.get_model_join(obj):
                        out[field_name] = self.serialize_queryset(v, referrers + (obj,))
                    else:
                        out[field_name] = [o.pk for o in v]
                else:
                    out[field_name] = []
        except Exception as exc:  # pragma: no cover
            logger.error(f"serialize_many_relationship error - {obj}", exc_info=exc)
        return out

    def serialize_value_field(self, obj: models.Model, field: Any) -> Dict[Any, Any]:
        """
        Serializes regular 'jsonable' field (Char, Int, etc.) of Django model instance
        """
        return {field.name: getattr(obj, field.name)}

    def serialize_data(self, data: Any) -> Any:
        out = data
        # Queryset
        if self.is_queryset(data):
            out = self.serialize_queryset(data)
        # Model
        elif self.is_model_instance(data):
            out = self.serialize_model_instance(data)
        # Add limit_off pagination support
        elif self.is_paginated(data):
            out = self.serialize_queryset(data.get("items"))
        return out


django_serializer = DjangoSerializer()



import logging
from typing import Any, Dict, List, Tuple, Type

from django.db import models, transaction
from django.db.models.query import QuerySet
from ninja_extra.shortcuts import get_object_or_none

from easy.exception import BaseAPIException

logger = logging.getLogger(__name__)


class CrudModel(object):
    def __init__(self, model: Type[models.Model]):
        self.model = model
        self.m2m_fields_list: List = list(
            _field
            for _field in self.model._meta.get_fields(include_hidden=True)
            if isinstance(_field, models.ManyToManyField)
        )

    def _separate_payload(self, payload: Dict) -> Tuple[Dict, Dict]:
        m2m_fields = {}
        local_fields = {}
        for _field in payload.keys():
            model_field = self.model._meta.get_field(_field)
            if model_field in self.m2m_fields_list:
                m2m_fields.update({_field: payload[_field]})
            else:
                # Handling FK fields ( append _id in the end)
                if self.model._meta.get_field(_field).is_relation:
                    if not f"{_field}".endswith("_id"):
                        local_fields.update({f"{_field}_id": payload[_field]})
                else:
                    local_fields.update({_field: payload[_field]})
        return local_fields, m2m_fields

    @staticmethod
    def _crud_set_m2m_obj(obj: models.Model, m2m_fields: Dict) -> None:
        if obj and m2m_fields:
            for _field, _value in m2m_fields.items():
                if _value and isinstance(_value, List):
                    m2m_f = getattr(obj, _field)
                    m2m_f.set(_value)

    # Define BASE CRUD
    @transaction.atomic()
    def _crud_add_obj(self, **payload: Dict) -> Any:
        local_f_payload, m2m_f_payload = self._separate_payload(payload)

        try:
            # Create obj with local_fields payload
            obj = self.model.objects.create(**local_f_payload)
            # Save obj with m2m_fields payload
            self._crud_set_m2m_obj(obj, m2m_f_payload)
        except Exception as e:  # pragma: no cover
            raise BaseAPIException(f"Create Error - {e}")
        if obj:
            return obj.id

    def _crud_del_obj(self, pk: int) -> bool:
        obj = get_object_or_none(self.model, pk=pk)
        if obj:
            self.model.objects.filter(pk=pk).delete()
            return True
        else:
            return False

    @transaction.atomic()
    def _crud_update_obj(self, pk: int, payload: Dict) -> bool:
        local_fields, m2m_fields = self._separate_payload(payload)
        if not self.model.objects.filter(pk=pk).exists():
            return False
        try:
            obj, _ = self.model.objects.update_or_create(pk=pk, defaults=local_fields)
            self._crud_set_m2m_obj(obj, m2m_fields)
        except Exception as e:  # pragma: no cover
            raise BaseAPIException(f"Update Error - {e}")
        return bool(obj)

    def _crud_get_obj(self, pk: int) -> Any:
        if self.m2m_fields_list:
            qs = self.model.objects.filter(pk=pk).prefetch_related(
                self.m2m_fields_list[0].name
            )
            for f in self.m2m_fields_list[1:]:
                qs = qs.prefetch_related(f.name)
        else:
            qs = self.model.objects.filter(pk=pk)
        if qs:
            return qs.first()

    def _crud_get_objs_all(self, **filters: Any) -> Any:
        """
        CRUD: get multiple objects, with django orm filters support
        Args:
            filters: {"field_name__lte", 1}
        Returns: qs

        """
        qs = None
        if filters:
            try:
                qs = self.model.objects.filter(**filters)
            except Exception as e:  # pragma: no cover
                logger.error(e)
        else:
            qs = self.model.objects.all()
        # If there are 2m2_fields
        if self.m2m_fields_list and qs:
            qs = qs.prefetch_related(self.m2m_fields_list[0].name)
            for f in self.m2m_fields_list[1:]:
                qs = qs.prefetch_related(f.name)
        return qs

    def _crud_filter(self, **kwargs: Any) -> QuerySet:
        return self.model.objects.filter(**kwargs)  # pragma: no cover

    def _crud_filter_exclude(self, **kwargs: Any) -> QuerySet:
        return self.model.objects.all().exclude(**kwargs)


class BaseOrm(CrudModel):
    ...



from typing import TYPE_CHECKING, Any

from django.http import HttpRequest

from .domain import EventBiz

if TYPE_CHECKING:
    from ninja_extra.controllers.base import ControllerBase  # pragma: no cover

from easy.services import BaseService


class EventService(BaseService):
    def __init__(self, biz=EventBiz()):
        super().__init__(biz.model)

    @staticmethod
    async def prepare_create_event_data(data):
        """Helper func for unit testing"""
        object_data = data.copy()
        object_data.update(title=f"{object_data['title']}_create")
        return object_data

    async def get_event_objs_demo(self):
        """Demo API for unit testing"""
        return await self.get_objs()

    async def get_identity_demo(self, word):
        """Demo API for unit testing"""
        return dict(says=word)

    def check_permission(
        self, request: HttpRequest, controller: "ControllerBase"
    ) -> bool:
        """
        Overwrite parent check_permission
        """
        user = request.user
        return bool(
            user and user.is_authenticated and user.is_active
        ) and super().check_permission(request, controller)

    def check_object_permission(
        self, request: HttpRequest, controller: "ControllerBase", obj: Any
    ) -> bool:
        """
        Only superuser is granted access
        """
        return bool(request.user.is_superuser) and super().check_object_permission(
            request, controller, obj
        )



import json
from datetime import datetime, timedelta

import django
import pytest
from asgiref.sync import sync_to_async

from .easy_app.controllers import EasyCrudAPIController
from .easy_app.models import Event, Type
from .easy_app.schema import EventSchema
from .easy_app.services import EventService

dummy_data = dict(
    title="AsyncAPIEvent",
    start_date=str(datetime.now().date()),
    end_date=str((datetime.now() + timedelta(days=5)).date()),
)


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
@pytest.mark.django_db
class TestEasyCrudAPIController:
    async def test_crud_apis(self, transactional_db, easy_api_client):
        client = easy_api_client(EasyCrudAPIController)

        object_data = await EventService.prepare_create_event_data(dummy_data)

        response = await client.put(
            "/", json=object_data, content_type="application/json"
        )
        assert response.status_code == 200

        assert response.json().get("code") == 201
        event_id = response.json().get("data")["id"]

        response = await client.get(
            f"/{event_id}",
        )
        assert response.status_code == 200
        assert response.json().get("data")["title"] == "AsyncAPIEvent_create"

        response = await client.get("/")
        assert response.status_code == 200
        assert response.json().get("data")[0]["title"] == "AsyncAPIEvent_create"

    async def test_base_response(self, transactional_db, easy_api_client):
        client = easy_api_client(EasyCrudAPIController)

        response = await client.get(
            "/base_response/",
        )
        assert response.status_code == 200
        assert response.json().get("data")["data"] == "This is a BaseApiResponse."

    async def test_qs_paginated(self, transactional_db, easy_api_client):
        client = easy_api_client(EasyCrudAPIController)

        object_data = await EventService.prepare_create_event_data(dummy_data)

        response = await client.put(
            "/", json=object_data, content_type="application/json"
        )
        assert response.status_code == 200
        assert response.json().get("code") == 201
        event_id = response.json().get("data")["id"]
        response = await client.get(
            "/qs_paginated/",
        )
        assert response.status_code == 200
        assert response.json().get("data")[0]["id"] == event_id

    async def test_qs_list(self, transactional_db, easy_api_client):
        client = easy_api_client(EasyCrudAPIController)

        for i in range(4):
            type = await sync_to_async(Type.objects.create)(name=f"Test-Type-{i}")
            object_data = await EventService.prepare_create_event_data(dummy_data)
            object_data.update(
                title=f"{object_data['title']}_qs_list_{i}", type=type.id
            )
            await client.put("/", json=object_data)

        type = await sync_to_async(Type.objects.create)(name="Test-Type-88")
        object_data = await EventService.prepare_create_event_data(dummy_data)
        object_data.update(title=f"{object_data['title']}_qs_list_88", type=type)
        event = await sync_to_async(Event.objects.create)(**object_data)

        response = await client.get(
            "/qs_list/",
        )
        assert response.status_code == 200

        data = response.json().get("data")

        assert data[0]["title"] == "AsyncAPIEvent_create_qs_list_0"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert (
            event_schema["title"]
            == data[4]["title"]
            == "AsyncAPIEvent_create_qs_list_88"
        )

    async def test_qs_(self, transactional_db, easy_api_client):
        client = easy_api_client(EasyCrudAPIController)

        for i in range(4):
            type = await sync_to_async(Type.objects.create)(name=f"Test-Type-{i}")
            object_data = await EventService.prepare_create_event_data(dummy_data)
            object_data.update(title=f"{object_data['title']}_qs_{i}", type=type.id)
            await client.put("/", json=object_data)

        response = await client.get(
            "/qs/",
        )
        assert response.status_code == 200

        assert response.json().get("data")[0]["title"] == "AsyncAPIEvent_create_qs_0"



import json
from typing import Any, Type

from django.db.models.fields.files import FieldFile, ImageFieldFile
from ninja.renderers import JSONRenderer
from ninja.responses import NinjaJSONEncoder


class EasyJSONEncoder(NinjaJSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, ImageFieldFile) or isinstance(o, FieldFile):
            try:
                return o.path
            except NotImplementedError:
                return o.url or o.name  # pragma: no cover
            except ValueError:
                return ""

        return super().default(o)


class EasyJSONRenderer(JSONRenderer):
    encoder_class: Type[json.JSONEncoder] = EasyJSONEncoder



from unittest.mock import Mock

from ninja.openapi.urls import get_openapi_urls

from easy import EasyAPI
from easy.decorators import docs_permission_required


def test_docs_decorator_staff_members():
    api = EasyAPI(docs_decorator=docs_permission_required)
    paths = get_openapi_urls(api)

    assert len(paths) == 2
    for ptrn in paths:
        request = Mock(user=Mock(is_staff=True))
        result = ptrn.callback(request)
        assert result.status_code == 200

        request = Mock(user=Mock(is_staff=False))
        request.build_absolute_uri = lambda: "http://example.com"
        result = ptrn.callback(request)
        assert result.status_code == 302
